# -*- coding: utf-8 -*-

# Standard Library
import json
from collections import defaultdict
# Vendor Library
from django.db.transaction import commit_on_success

# Project Library
from kinder.core.declaration.models import Declaration
from kinder.core.declaration_status.models import (
    DeclarationStatus)
from kinder.core.queue_module.api import get_queue_by_context
from kinder.core.queue_module.context import QueueContext

from kinder.core.audit_log.models import AdditionalEncoder
from kinder.core.audit_log.helpers import get_model
from kinder.core.children.models import ChildrenDelegate
from concentrator.change import (
    StorageHelper, DeclarationChangeHelper)
from concentrator.models import ChangeDeclaration
from concentrator.rules import *
from concentrator.webservice.spyne_objects_proxy import *
from concentrator.webservice.django_objects_proxy import *
from concentrator.webservice.helpers import (
    get_delegate, DocumentWorker, _get_document_reference_from_ctx)
from concentrator import settings
from kinder.provider import features


class NewApplicationProxy(object):
    """
    Прокси объект для метода NewApplication
    """

    def __init__(self, request, binary_data, req_code):

        self._external_id = request.ExternalId
        self._submit_date = request.SubmitDate
        self._entry_date = request.EntryDate
        self._adaptation_program_type = request.AdaptationProgramType
        self._delegate_data = request.Applicant
        self._child_data = request.DeclaredPerson
        self._edu_organizations = request.EduOrganizations
        self._binary_data = binary_data
        self._request_code = req_code
        self._document_references = _get_document_reference_from_ctx(
            request)

    def _create_delegate(self):

        return DelegateProxy(self._delegate_data).get()

    def _create_children(self):

        return ChildrenProxy(
            self._child_data, self._delegate_data,
            self._adaptation_program_type).get()

    def _create_benefits(self):
        result = []
        doc_number = (self._child_data.Benefits.BenefitsDocInfo
                      if self._child_data.Benefits else None)
        if self._child_data.Benefits and self._child_data.Benefits.Benefit:
            result = PrivilegeProxy([
                (benefit.name,
                 benefit.Type)
                for benefit in self._child_data.Benefits.Benefit]).get()
        return result, doc_number

    def _create_declaration(self):
        return Declaration(
            date=self._submit_date,
            desired_date=self._entry_date,
            offer_other=self._edu_organizations.AllowOfferOther,
            client_id=self._external_id,
            source=features.get_feature(
                'concentrator.webservice.service.'
                'NewApplicationRequest.get_source')(),
            portal=1)

    def _attache_documents(self, declaration):

        DocumentWorker(
            declaration, self._binary_data, self._request_code).attach()

    def _save_doc_extra_info(self, declaration):

        DocExtraInfoProxy(declaration, self._document_references).save()

    def _save_privileges(self, privileges, declaration, doc_number, delegate):
        for privilege in privileges:
            DeclarationPrivilege.objects.create(
                declaration=declaration,
                datetime=datetime.datetime.now(),
                privilege=privilege,
                doc_number=doc_number)

    @commit_on_success
    def process(self):
        if Declaration.objects.filter(client_id=self._external_id).exists():
            raise SpyneException(
                u'Заявление с ExternalId = %s уже существует' %
                self._external_id)
        declaration = self._create_declaration()
        delegate, delegate_extension = self._create_delegate()
        children = self._create_children()
        privileges, doc_number = self._create_benefits()
        mo = get_mo_from_request(self._edu_organizations, declaration)

        delegate.save()
        delegate_extension.delegate = delegate
        delegate_extension.save()
        children.save()
        ChildrenDelegate.objects.create(
            children=children, delegate=delegate)

        declaration.children = children
        declaration.mo = mo
        declaration.status = DeclarationStatus.objects.get(
            code=DSS.DUL_CONFIRMATING)
        declaration.save()

        declaration_units = get_declaration_units_from_request(
            self._edu_organizations, declaration)

        for declaration_unit in declaration_units:
            # bulk create не использую т.к. он не вызывает save
            # метода и не срабатывают сигналы post_save и pre_save
            declaration_unit.declaration = declaration
            declaration_unit.save()

        self._save_privileges(privileges, declaration, doc_number, delegate)
        self._save_doc_extra_info(declaration)
        self._attache_documents(declaration)

        return declaration.id


class GetApplicationQueueProxy(object):
    """
    Прокси объект для метода GetApplicationQueue
        Запрос текущей очереди заявления
    Возвращает список очередей во все желаемые доу заявки,
    каждая очередь содержит все заявки до нашей.
    """
    def __init__(self, declaration, request):
        assert declaration
        assert isinstance(declaration, Declaration)
        self.declaration = declaration
        self.request = request

    @staticmethod
    def _create_application(declaration, i):
        return ApplicationProxy.create_application(declaration, i)

    @staticmethod
    def _create_queue(unit, count, app):
        conf = {
            'EduOrganizationCode': str(unit.id),
            'ApplicationsCount': count,
            'Application': app
        }
        q = Queue(**conf)
        return q

    def process(self):
        """
        1. Если AllApplications == false, значит это первый запрос и мы должны
        отдать информацию по текущему учреждению, и вернуть им признак
        разрешено у нас отдавать очередь или нет в тэге SupportAllApplications
        2. Если передано AllApplications == true, но у нас в настроках
        SupportAllApplications = false,
         то отрабатываем как при AllApplications == false
        3. Если передано AllApplications == true и у нас в настроках
        SupportAllApplications = true То отдаем всю очередь перед заявкой
        4. Если передан id учреждения, то фильтруем желаемые учреждения по нему
        """
        result = []
        decl_id = self.declaration.id
        all_app = self.request.AllApplications
        if all_app and not settings.GET_APPQ_FULL:
            all_app = False

        if self.declaration.status.code not in DSS.status_queue():
            return GetApplicationQueueResponse(
                Queue=result, SupportAllApplications=settings.GET_APPQ_FULL)

        qs = self.declaration.declarationunit_set.all()
        if self.request.EduOrganizationCode:
            qs = qs.filter(unit_id=self.request.EduOrganizationCode)
        qs = qs.order_by('ord')

        for d_unit in qs:
            unit = d_unit.unit
            # считаем очередь до заявки в желаемое доу
            ctx = QueueContext.make_base_context(
                self.declaration.children.date_of_birth,
                unit)
            app = []
            i = 0
            for i, decl in enumerate(get_queue_by_context(ctx), 1):
                if not all_app and decl_id != decl.id:
                    continue
                app.append(
                    GetApplicationQueueProxy.
                    _create_application(decl, i))
                if decl.id == decl_id:
                    break
            result.append(
                GetApplicationQueueProxy._create_queue(
                    unit, i, app))

        return GetApplicationQueueResponse(
            Queue=result, SupportAllApplications=settings.GET_APPQ_FULL)


class GetApplicationStateProxy(object):
    """
    Прокси объект для метода GetApplicationQueue
        Запрос текущего статуса заявления
    """
    def __init__(self, declaration):
        assert declaration
        assert isinstance(declaration, Declaration)
        self.declaration = declaration

    def process(self):
        status_code = DeclarationStatusCodeRule.get_concetr(
            self.declaration.status.code)
        return GetApplicationStateResponse(**{
            'Code': str(status_code) if status_code else None,
            'Name': DeclarationStatusNameRule.get_concetr(
                self.declaration.status.code),
            'Details': self.declaration.status.description
        })


class FindApplicationsProxy(object):

    BY_DELEGATE = 1
    BY_CHILD = 2

    DOC_TYPE_RULE = {
        BY_DELEGATE: DelegateDocTypeRule,
        BY_CHILD: ChildrenDocTypeRule
    }

    def __init__(self, person_info, mode=BY_DELEGATE):

        self._person = person_info
        self._search_dict = {}
        self._mode = mode

    def _fill_attribute(
            self, spyne_attribute, django_field):

        value = getattr(self._person, spyne_attribute, None)
        if value:
            self._search_dict[django_field] = value

    def _convert_doc_type(self):

        value = getattr(self._person, 'DocType', None)
        if value:
            self._person.DocType = self.DOC_TYPE_RULE[
                self._mode].get_system(
                    self._person.DocType)

    def _convert_data(self):

        self._convert_doc_type()

    def _find_declarations(self, prefix):

        self._convert_data()

        self._fill_attribute('FirstName', '%s__firstname' % prefix)
        self._fill_attribute('LastName', '%s__surname' % prefix)
        self._fill_attribute('MiddleName', '%s__patronymic' % prefix)
        self._fill_attribute('Snils', '%s__snils' % prefix)
        self._fill_attribute('DateOfBirth', '%s__date_of_birth' % prefix)
        self._fill_attribute('DocType', '%s__dul_type' % prefix)
        self._fill_attribute('DocNumber', '%s__dul_number' % prefix)
        self._fill_attribute('DocIssueDate', '%s__dul_date' % prefix)

        declarations = Declaration.objects.filter(
            **self._search_dict).distinct('id')
        return declarations

    def process(self):
        if self._mode == self.BY_CHILD:
            type_response = FindApplicationsByDeclaredPersonResponse
            declarations = self._find_declarations('children')
        else:
            type_response = FindApplicationsByApplicantResponse
            declarations = self._find_declarations(
                'children__childrendelegate__delegate')

        applicants = []
        for declaration in declarations:

            applicants.append(
                ApplicationSearchResultProxy.create_application(
                    declaration))

        return type_response(Application=applicants)


class GetApplicationProxy(object):
    """
    Прокси объект для метода GetApplication
        Получение данных Заявления для изменения
    """
    def __init__(self, declaration, decl_proxy, app_proxy, dou_proxy):
        assert declaration
        assert isinstance(declaration, Declaration)
        self.declaration = declaration
        self.decl_proxy = decl_proxy
        self.app_proxy = app_proxy
        self.dou_proxy = dou_proxy

    def process(self):
        conf = {
            'SubmitDate': self.declaration.date,
            'EntryDate': get_desire_date(self.declaration),
            'EducationProgramType': '',
            'AdaptationProgramType': get_health_need(self.declaration),
            'Applicant': self.app_proxy.create_spyne_obj(),
            'DeclaredPerson': self.decl_proxy.create_spyne_obj(),
            'EduOrganizationsData': self.dou_proxy.create_spyne_obj(),
            'ApplicationRules': ApplicationRulesDataProxy.create_rules_data(
                self.declaration),
            'DocumentReferences': DocumentReferencesDataProxy.get_data(
                self.declaration)
        }

        return GetApplicationResponse(**conf)


class UpdateProxy(NewApplicationProxy):
    """
    Прокси объект отвечает за сохранение запроса
    об изменении заявки
    """

    def __init__(self, request, declaration, binary_data, request_code):
        self._external_id = request.ExternalId
        self._submit_date = request.SubmitDate
        self._entry_date = request.EntryDate
        self._adaptation_program_type = request.AdaptationProgramType
        self._delegate_data = request.Applicant
        self._child_data = request.DeclaredPerson
        self._edu_organizations = request.EduOrganizationsData
        self.declaration = declaration
        self.result = []
        self._binary_data = binary_data
        self._request_code = request_code
        self._document_references = _get_document_reference_from_ctx(
            request)

    def chek_model(self, name_model, conc_data, have_data):
        """
        @param name_model: str имя модели из MAP_CHANGES
        @param conc_data: spyne object инстанс модеи воссозданный
         из данных на изменение
        @param have_data: django model инстанс моделииз хранилища
        """
        class_model = get_model(name_model)
        class_helper = StorageHelper.get_change_helper(name_model)
        ch_help = class_helper(class_model, name_model)
        ch_help.check_diff(conc_data, have_data)
        self.result.append(ch_help)

    def _get_declaration_units_list(self):
        """Возвращает список (unit_id, ord) по данным из запроса,
        прошедших всю валидацию"""
        decl_units = get_declaration_units_from_request(
            self._edu_organizations, self.declaration)
        return map(lambda x: (x.unit_id, x.ord), decl_units)

    def _create_benefits(self):
        result, doc_number = super(UpdateProxy, self)._create_benefits()
        return map(lambda x: x.id, result)

    def _save_documents(self):
        """
        Сохранение документов
        """
        DocumentWorker(
            self.declaration, self._binary_data, self._request_code).attach()
        DocExtraInfoProxy(self.declaration, self._document_references).save()

    def _check_benefits(self):
        conc_benefits = self._create_benefits()
        benefits = DeclarationPrivilege.objects.filter(
            declaration=self.declaration).values_list('privilege_id', flat=True)
        self.chek_model('DeclarationPrivilege', conc_benefits, benefits)

    def process(self):
        msg = u'OK'
        #проверем поля заявки
        self.chek_model('Declaration', self._create_declaration(),
                        self.declaration)
        #проверяем поля ребенка
        self.chek_model('Children', self._create_children(),
                        self.declaration.children)
        #проверяем поля представителя
        conc_delegate, _ = self._create_delegate()
        #тут также надеемся что родитль один и его же мы и редактируем!
        delegate = get_delegate(self.declaration)
        self.chek_model('Delegate', conc_delegate, delegate)

        # проверяем поле ЖУ
        conc_decl_units = self._get_declaration_units_list()
        sys_decl_units = get_declaration_units(self.declaration)
        self.chek_model('DeclarationUnit', conc_decl_units, sys_decl_units)

        self._check_benefits()

        result = StorageHelper.create_change(self.declaration, self.result)
        if not result:
            msg = u'Diff not found'

        # Сохранение документов без подтверждения
        self._save_documents()

        return UpdateApplicationResponse(**{
            'Status': msg
        })


class ResetDeclarationProxy(object):

    STATUS_TRANSFER_COMMENT = u"""
        Авто-смена статуса в связи с поступившей заявкой
        на отказ от участия в очереди поступила с ЕПГУ.
        Необходимо уточнить актуальность запроса у законных
        представителей.
        """

    def __init__(self, request, declaration, **kwargs):

        self._request = request
        self.declaration = declaration
        self.result = defaultdict(list)

    def process(self):
        ch_help = DeclarationChangeHelper(Declaration, 'Declaration')
        self.result['Declaration'].extend(
            ch_help.get_status_change(self.declaration))

        ChangeDeclaration(
            declaration=self.declaration,
            data=json.dumps(
                self.result, cls=AdditionalEncoder)).save()

        return UpdateApplicationResponse(**{
            'Status': u'Успех'
        })


class UpdateApplicationProxy(object):

    RESET_STATE = u'11'

    DEFAULT_BEHAVIOR_CLS = UpdateProxy

    BEHAVIOR = {
        RESET_STATE: ResetDeclarationProxy
    }

    @classmethod
    def process(cls, request, declaration, binary_data, request_code):

        return cls.BEHAVIOR.get(
            request.State,
            cls.DEFAULT_BEHAVIOR_CLS)(
                request, declaration,
                binary_data=binary_data, request_code=request_code).process()
